/*
 * Copyright (C) 1997
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cbqd.c,v 0.5 1998/04/03 07:34:23 kjc Exp kjc $
 */
/*******************************************************************

  Copyright (c) 1996 by the University of Southern California
  All rights reserved.

  Permission to use, copy, modify, and distribute this software and its
  documentation in source and binary forms for any purpose and without
  fee is hereby granted, provided that both the above copyright notice
  and this permission notice appear in all copies. and that any
  documentation, advertising materials, and other materials related to
  such distribution and use acknowledge that the software was developed
  in part by the University of Southern California, Information
  Sciences Institute.  The name of the University may not be used to
  endorse or promote products derived from this software without
  specific prior written permission.

  THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
  the suitability of this software for any purpose.  THIS SOFTWARE IS
  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  Other copyrights might apply to parts of this software and are so
  noted when applicable.

********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <err.h>

#include <sys/socket.h>

#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <netinet/altq_cbq.h>
#include "rsvp_cbq.h"
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

/* from rsvp_main.c */
char *solconfigfile = "/etc/cbq.conf";
#define  CBQD_PID_FILE	  "/var/tmp/cbqd.pid"

/* from rsvp_global.h */
int	if_num;		/* number of phyints */
int	rsvp_errno;	/* error code -- used to communicate
			 *	error from kernel interface
			 */
int	debug;		/* Debug output control bits */
int	l_debug;	/* Logging severity level */
int	Debug_mode;	/* nosched (dummy mode) */

int daemonize = 1;

/* for command mode */
int             T;  		/* Current Thread number */
FILE		*infp;		/* Input file pointer */
char           *infile = NULL;  /* command input file.  stdin if NULL. */
fd_set		fds, t_fds;

#define  DEFAULT_DEBUG_MASK	0	
#define  DEFAULT_LOGGING_LEVEL	LOG_INFO

void usage(void)
{
	fprintf(stderr, "usage: cbqd [options]\n");
	fprintf(stderr, "    options:\n");
	fprintf(stderr, "    -f config_file	: set config file\n");
	fprintf(stderr, "    -v			: verbose (no daemonize)\n");
	fprintf(stderr, "    -d			: debug (no daemonize)\n");
}

static void
sigint_handler(int sig)
{
	cbq_destroyall();

	if (sig == SIGHUP) {
		printf("reinitializing cbq...\n");
		cbq_init();
		return;
	}

	fprintf(stderr, "Exiting on signal %d\n", sig);
	/* if we have a pid file, remove it */
	if (daemonize)
		unlink(CBQD_PID_FILE);
	exit(0);
}

int main(int argc, char **argv)
{
	u_int           c;
	extern char    *optarg;

	debug = DEFAULT_DEBUG_MASK;
	l_debug = DEFAULT_LOGGING_LEVEL;

	while ((c = getopt(argc, argv, "f:vDdl:")) != -1) {
		switch (c) {
		case 'f':
			solconfigfile = optarg;
			break;

		case 'D':	/* -D => dummy mode */
			Debug_mode = 1;
			printf("Debug mode set.\n");
			break;
			
		case 'v':
			l_debug = LOG_DEBUG;
			debug |= DEBUG_CBQ;
			daemonize = 0;
			break;

		case 'd':
			daemonize = 0;
			break;

		case 'l':
			l_debug = atoi(optarg);
			break;

		default:
			usage();
		}
	}

	signal(SIGINT, sigint_handler);
	signal(SIGTERM, sigint_handler);
	signal(SIGHUP, sigint_handler);

	if (daemonize)
		openlog("fifoqd", LOG_PID | LOG_PERROR, LOG_DAEMON);

	cbq_init();

	if (daemonize) {
		FILE *fp;

		printf("cbqd: going into the daemon mode.\n");

		daemon(0, 0);

		/* save pid to the pid file (/var/tmp/cbqd.pid) */
		if ((fp = fopen(CBQD_PID_FILE, "w")) != NULL) {
			fprintf(fp, "%d\n", getpid());
			fclose(fp);
		}
		else
			warn("can't open pid file: %s: %s",
			     CBQD_PID_FILE, strerror(errno));
	    
		/*
		 * if we are in the daemon mode, just do nothing for the
		 * rest of our life but keeping cbq device open...
		 */
		while (1) {
			pause();
		}
	}

	if (infile) {
		if (NULL == (infp = fopen(infile, "r"))) {
			perror("Cannot open input file");
			exit(1);
		}
	} else {
		infp = stdin;
		printf("\nEnter ? or command:\n");
		printf("cbq %s> ", cur_ifname());
	}

	/*
	 * go into the command mode.
	 * the code below is taken from rtap of rsvpd.
	 */
	/* rtap basically uses standard I/O.  However, in order to
	 *	receive asynchronous events from RAPI, it must do a
	 *	select() call on its TTY input and on the API fd.
	 */
	
	fflush(stdout);
	FD_ZERO(&fds);
	while (1) {
		int rc;

		FD_COPY(&fds, &t_fds);
		FD_SET(fileno(infp), &t_fds);

		rc = select(FD_SETSIZE, &t_fds, NULL, NULL, NULL);
		if (rc < 0) {
			if (errno != EINTR) {	
				perror("select");
				exit(1);
			}
			continue;
		}
		
		/*	If there is control input, read the input line,
		 *	parse it, and execute.
		 */
		if (FD_ISSET(fileno(infp), &t_fds)) {
			rc = DoCommand(infile, infp);
			if (rc == -1) {
				/* Closed API pipe
				 */
				printf("cbq %s> ", cur_ifname());
			}
			else if (rc == 0) {
				/*
				 *	EOF on input.  If reading from file,
				 *	go to stdin; else exit.
				 */
				if (infile) {
					infp = stdin;
					infile = NULL;
					printf("\nEnter ? or command:\n");
				} else
					exit(0);
			}
			else if (!infile)
				printf("cbq %s> ", cur_ifname());
			fflush(stdout);
		}
	}
}

/* taken from rsvp_debug.c and modified. */
void
log(int severity, int syserr, const char *format, ...)
{
	va_list ap;

#ifdef __STDC__
	va_start(ap, format);
#else
	va_start(ap);
#endif

	if (severity <= l_debug) {
		if (!daemonize) {
			vfprintf(stderr, format, ap);
			fprintf(stderr, "\n");
		}
		else
			vsyslog(severity, format, ap);
	}

	va_end(ap);

	if (syserr == 0) {
		/* Do nothing for now */
	} else if (syserr < sys_nerr) {
		if (severity <= l_debug) {
			if (!daemonize)
				fprintf(stderr, ": %s\n", sys_errlist[syserr]);
			else
				syslog(severity, ": %s", sys_errlist[syserr]);
		}
	} else {
		if (severity <= l_debug) {
			if (!daemonize)
				fprintf(stderr, ": errno %d\n", syserr);
			else
				syslog(severity, ": errno %d\n", syserr);
		}
	}
}

