/* $Id: viewer.c,v 0.2 1998/04/03 07:33:30 kjc Exp kjc $ */
/*
 *  Copyright (c) 1996
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms of parts of or the
 * whole original or derived work are permitted provided that the above
 * copyright notice is retained and the original work is properly
 * attributed to the author. The name of the author may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
/* viewer.c -- a viewer module for remote-monitoring.  this module is
   shared by tttview and ttttextview but TTT_TEXT flag is set for
   ttttextview. */
#include <stdio.h>
#include <netdb.h>
#include <limits.h>
#include <string.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "ttt.h"
#include "ttt_node.h"
#include "ttt_remote.h"

#include "cbqmon.h"

#define  PID_FILE	  "./monitor.pid"

#define BUFFER_SIZE	4096	/* big enough */
static char buffer[BUFFER_SIZE];	/* receive buffer */
static char *class_name_tab[CBQM_MAXCLASS];

static struct sockaddr_in probe_addr;
static char *probe_name = NULL;
static int packets_received;
static struct timeval remote_time;
static int shared_port = 1;		/* share the multicast port */
static int multicast = 0;		/* use multicast */

static char *dumpfile = NULL;
static FILE *dumpf = NULL;
static char *configfile = NULL;
static char *replayfile = NULL;
static FILE *replayf = NULL;
static int replayspeed = 1;
static int create_pidfile = 0;
static int show_root = 0;
static int got_sighup = 0;

void view_sockread(int clientdata, int mask);
static int read_replay(FILE *df, char *buf);
static int read_cinfo(struct class_info *cinfo);
double get_remotetime(void);
static void endian_swap(char *buf, int size);
static void sighup_handler(int sig);
static int next_word(char **cpp, char *b);
static int get_ifname(char **cpp, char **ifnamep);
static int read_config(char *configfile);
static char *cbq_class_cmd(char *cmdbuf) ;

static void usage()
{
    printf("usage: cbqmonitor [options]\n");
    printf(" options:\n");
    printf("	[-multicast]\n");
    printf("    [-addr recv_addr]\n");
    printf("    [-mcastifaddr mcast_if_addr]\n");
    printf("	[-port recv_port]\n");
    printf("	[-probe probe_addr]\n");
    printf("	[-config config_file]\n");
    printf("	[-dump filename]\n");
    printf("	[-replay filename [-speed #]]\n");
    printf("	[-pid]\n");
    printf("	[-showroot]\n");
    exit(1);
}

void view_parseargs(int argc, char **argv)
{
    int i;

    for (i=1; i<argc; i++)
	if (strncmp(argv[i], "-multicast", 4) == 0)
	    ttt_viewname = TTT_MCASTADDR;
	else if (strncmp(argv[i], "-addr", 4) == 0 && ++i < argc)
	    ttt_viewname = argv[i];
	else if (strncmp(argv[i], "-mcastifaddr", 4) == 0 && ++i < argc)
	    ttt_mcastif = argv[i];
	else if (strncmp(argv[i], "-port", 4) == 0 && ++i < argc)
	    ttt_portno = atoi(argv[i]);
        else if (strncmp(argv[i], "-probe", 4) == 0 && ++i < argc)
	    probe_name = argv[i];
	else if (strncmp(argv[i], "-dump", 4) == 0 && ++i < argc)
	    dumpfile = argv[i];
	else if (strncmp(argv[i], "-config", 4) == 0 && ++i < argc)
	    configfile = argv[i];
	else if (strncmp(argv[i], "-replay", 4) == 0 && ++i < argc)
	    replayfile = argv[i];
	else if (strncmp(argv[i], "-speed", 4) == 0 && ++i < argc)
	    replayspeed = atoi(argv[i]);
	else if (strncmp(argv[i], "-pid", 4) == 0)
	    create_pidfile = 1;
	else if (strncmp(argv[i], "-showroot", 4) == 0)
	    show_root = 1;
	else if (strncmp(argv[i], "-help", 4) == 0)
	    usage();
	else if (strncmp(argv[i], "-version", 4) == 0) {
	    printf("%s\n", ttt_version);
	    exit(0);
	}
}

int view_opensock(void)
{
    char my_name[MAXHOSTNAMELEN], class_name[64];
    struct sockaddr_in my_addr;
    int sockfd, class_no;
    FILE *fp;

    signal(SIGHUP, sighup_handler);

    /* if config file is specified, read class names from the config file */
    if (configfile != NULL)
	read_config(configfile);
    else {
	/* initialize class names */
	for (class_no = 0; class_no < CBQM_MAXCLASS; class_no++) {
	    sprintf(class_name, "class-%d", class_no);
	    class_name_tab[class_no] = strdup(class_name);
	}
    }
    
    if (replayfile != NULL) {
	/* if replayfile is specified, open dumpfile */
	if ((replayf = fopen(replayfile, "r")) == NULL)
	    fatal_error("can't open replay file %s!", replayfile);
	return (TTT_TIMER_HANDLER);
    }

    if (create_pidfile)
	if ((fp = fopen(PID_FILE, "w")) != NULL) {
	    fprintf(fp, "%d\n", getpid());
	    fclose(fp);
	}
    
    if (ttt_viewname == NULL) {
	/* use my host name */
	if (gethostname(my_name, MAXHOSTNAMELEN) == -1)
	    fatal_error("no host_name");
    }
    else
	strcpy(my_name, ttt_viewname);
    if (name2sockaddrin(my_name, ttt_portno, &my_addr) < 0)
	fatal_error("can't get my address!");

#ifdef IN_MULTICAST
    if (IN_MULTICAST(ntohl(my_addr.sin_addr.s_addr)))
	multicast = 1;
#endif
    if ((sockfd = socket(PF_INET, SOCK_DGRAM, 0)) < 0)
	fatal_error("receiver: can't open socket");

    if (multicast) {
#ifdef IP_ADD_MEMBERSHIP
	struct ip_mreq mreq;
	struct sockaddr_in ifaddr;

	mreq.imr_multiaddr.s_addr = my_addr.sin_addr.s_addr;
	if (ttt_mcastif == NULL) {
	    mreq.imr_interface.s_addr = htonl(INADDR_ANY); /* use default if */
	}
	else {
	    if (name2sockaddrin(ttt_mcastif, ttt_portno, &ifaddr) < 0)
		fatal_error("can't get local address!");
	    mreq.imr_interface.s_addr = ifaddr.sin_addr.s_addr;
	}

	if (setsockopt(sockfd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		       (char *)&mreq, sizeof(mreq)) < 0)
	    fatal_error("can't join group");
	
	printf("joined multicast group: %s\n", ttt_viewname);
#else
	fatal_error("IP multicat not supported!");
#endif /* IP_ADD_MEMBERSHIP */
    }

    if (shared_port) {
	int one = 1;
	if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
		       (char *)&one, sizeof(one)) < 0)
	    fatal_error("can't share the port");
	printf("port %d is shared.\n", ttt_portno);
    }

    if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(my_addr)) < 0)
	fatal_error("receiver: can't bind");

    if (!multicast)
	printf("viewer started. reading from %s:%d ....\n",
	       inet_ntoa(my_addr.sin_addr), ttt_portno);


    if (probe_name != NULL) {
	if (name2sockaddrin(probe_name, 0, &probe_addr) < 0)
	    fatal_error("can't get probe address!");
	printf("reading from [%s] ....\n", probe_name);
    }
    else
	bzero(&probe_addr, sizeof(probe_addr));

    if (dumpfile != NULL) {
	if ((dumpf = fopen(dumpfile, "w")) == NULL)
	    fatal_error("can't open dumpfile %s\n", dumpfile);
    }

    return sockfd;
}

void view_closesock(int sockfd)
{
    if (replayf != NULL) {
	fclose(replayf);
	return;
    }
#ifdef IP_ADD_MEMBERSHIP
    if (multicast) {
	char my_name[MAXHOSTNAMELEN];
	struct sockaddr_in my_addr;
	struct ip_mreq mreq;
	
	if (ttt_viewname == NULL) {
	    /* use my host name */
	    if (gethostname(my_name, MAXHOSTNAMELEN) == -1)
		fatal_error("no host_name");
	}
	else
	    strcpy(my_name, ttt_viewname);
	if (name2sockaddrin(my_name, ttt_portno, &my_addr) < 0)
	    fatal_error("can't get my address!");

	mreq.imr_multiaddr.s_addr = my_addr.sin_addr.s_addr;
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);

	if (setsockopt(sockfd, IPPROTO_IP, IP_DROP_MEMBERSHIP,
		       (char *)&mreq, sizeof(mreq)) < 0)
	    perror("can't drop group");
    }
#endif /* IP_ADD_MEMBERSHIP */
    close(sockfd);

    if (dumpf != NULL)
	fclose(dumpf);
    if (create_pidfile)
	unlink(PID_FILE);
}

void view_sockread(int clientdata, int mask)
{
    int sockfd, nbytes, fromlen, rsize, seq_no, nrecords, i;
    struct sockaddr_in from_addr;
    struct ttt_hdr *hdr;
    char *cp;
    struct cbqm_info *info;

    sockfd = (int)clientdata;
    
    if (replayf == NULL) {
	fromlen = sizeof(from_addr);
	if ((nbytes = recvfrom(sockfd, buffer, BUFFER_SIZE, 0,
			   (struct sockaddr *)&from_addr, &fromlen)) == -1) {
	    perror("recvfrom");
	    return;
	}

	if (probe_addr.sin_addr.s_addr == 0) {
	    probe_addr = from_addr;
	    printf("reading from probe [%s] ....\n",
		   inet_ntoa(from_addr.sin_addr));
	}

	/* is this the probe we are reading from? */
	if (from_addr.sin_addr.s_addr != probe_addr.sin_addr.s_addr) {
	    static int warned = 0;
	    if (!warned) {
		printf("[warning] there are multiple probes.\n");
		printf("\tignoring probe: %s\n",
		       inet_ntoa(from_addr.sin_addr));
		warned = 1;
	    }
	    return;
	}
    }
    else {
	/* read from the replay file */
	nbytes = read_replay(replayf, buffer);
    }
    
    if (nbytes < 16 || (nbytes & 3)) {
	printf("sockread: packet too short size=%d\n", nbytes);
	return;
    }

    info = (struct cbqm_info *)buffer;
    if (info->cbqm_magic != CBQM_MAGIC) {
	if (info->cbqm_magic == CBQM_CIGAM)
	    endian_swap(buffer, nbytes);
	else {
	    printf("bad magic! 0x%x\n", info->cbqm_magic);
	    return;
	}
    }

#if 1
    /* reset the graph if we receive out-of-range timestamp */
    do {
	int diff = info->cbqm_sec - remote_time.tv_sec;
	if ((diff < -5 || diff > 10) && remote_time.tv_sec != 0) {
	    printf("timestamp out-of-range! reset the graph\n");
	    packets_received = 0;
	}
    } while (0);
#endif /* 1 */

    nrecords = info->cbqm_len;
    remote_time.tv_sec = info->cbqm_sec;
    remote_time.tv_usec = info->cbqm_usec;

    cp = (char *)&info->cbqm_cinfo[0];
    for (i=0; i<nrecords; i++) {
	if ((rsize = read_cinfo((struct class_info *)cp)) < 0)
	    break;
	cp += rsize;
    }

#ifdef TTT_TEXT
    ttt_textview(packets_received);
#else
    /* call ttt_display to update the graph */
    ttt_display(packets_received);
#endif
    packets_received++;

    if (dumpf != NULL) {
	fwrite(buffer, nbytes, 1, dumpf);
    }

    if (got_sighup) {
	if (configfile != NULL)
	    read_config(configfile);

	packets_received = 0;
	    
	if (dumpf != NULL) {
	    /* we have received SIGHUP, reopen the dump file */
	    fclose(dumpf);
	    if ((dumpf = fopen(dumpfile, "w")) == NULL)
		fatal_error("can't open dumpfile %s\n", dumpfile);
	}
	got_sighup = 0;
	signal(SIGHUP, sighup_handler);
    } 
}

static int read_replay(FILE *df, char *buf)
{
    struct cbqm_info *info = (struct cbqm_info *)buf;
    int len, hdr_len, interval, rval;
    int swap = 0;
    static struct timeval last_time;

    hdr_len = sizeof(struct cbqm_info)
	- CBQM_MAXCLASS * sizeof(struct class_info);
    
    /* read the header */
    if (fread(buf, hdr_len, 1, df) != 1) {
	if (feof(df))
	    printf("end-of-dumpfile\n");
	else
	    printf("can't read dumpfile header!");
#ifdef TTT_TEXT
	exit(0);
#else
	return (0);
#endif
    }

    if (info->cbqm_magic != CBQM_MAGIC) {
	if (info->cbqm_magic == CBQM_CIGAM) {
	    swap = 1;
	    endian_swap(buf, hdr_len);
	}
	else {
	    fatal_error("bad magic! 0x%x != 0x%x",
			info->cbqm_magic, CBQM_MAGIC);
	}
    }
    
    /* read the rest of the record */
    if (fread(buf+hdr_len, sizeof(struct class_info),
	      info->cbqm_len, df)
	!= info->cbqm_len) {
	printf("can't read dump file!");
	return (0);
    }
    if (swap)
	endian_swap(buffer + hdr_len,
		    info->cbqm_len*sizeof(struct class_info));
	
    len = hdr_len + info->cbqm_len * sizeof(struct class_info);

    if (last_time.tv_sec == 0)
	/* first record */
	interval = 0;
    else {
	long us;
	interval = (info->cbqm_sec - last_time.tv_sec) * 1000;
	us = (info->cbqm_usec - last_time.tv_usec);
	while (us < 0) {
	    interval -= 1;
	    us += 1000;
	}
	interval += us / 1000;
    }

    last_time.tv_sec = info->cbqm_sec;
    last_time.tv_usec = info->cbqm_usec;

    if (replayspeed > 1)
	interval = interval /replayspeed;

#ifdef TTT_TEXT
    if (replayspeed > 1) {
	struct timeval wait_time;

	wait_time.tv_sec = interval / 1000;
	wait_time.tv_usec = (interval % 1000) * 1000;
	select(0, NULL, NULL, NULL, &wait_time);
    }
#else
    if (replayspeed > 1)
	ttt_settimer(interval, replayf);
    else
	/* set minimum time unit (1ms) */
	ttt_settimer(1, replayf);
#endif
    
    return (len);
}

/* write node info to ttt_record */
static int read_cinfo(struct class_info *cinfo)
{
    int size, rval;
    long type, id, handle;
    static int last_bytes[CBQM_MAXCLASS];
    static int last_dropbytes[CBQM_MAXCLASS];

    handle = *(long *)&cinfo->handle;
    if (handle < 0)
	id = CBQM_MAXCLASS + handle;
    else
	id = handle;

#if 1
    /* if this is the first packet, adjust saved values */
    if (packets_received == 0) {
	last_bytes[id] = cinfo->nbytes;
	last_dropbytes[id] = cinfo->drop_bytes;
    }
#endif
    
    size = cinfo->nbytes - last_bytes[id];
    if (size < 0) {
	/* if size is a large negative value, the counter must have
	   wrapped arnound.  otherwise, the counter at the server
	   must have got reset. */
	if (size < INT_MIN / 2)
	    size += INT_MAX;
	else
	    size = 0;
    }

    if (size > 0 && (show_root || handle != 0xfffffffe))
	node_record(TTTTYPE_IP, &handle, size);

    size = cinfo->drop_bytes - last_dropbytes[id];
    if (size < 0) {
	/* if size is a large negative value, the counter must have
	   wrapped arnound.  otherwise, the counter at the server
	   must have got reset. */
	if (size < INT_MIN / 2)
	    size += INT_MAX;
	else
	    size = 0;
    }

    if (size > 0 && (show_root || handle != 0xfffffffe))
	node_record(TTTTYPE_IPHOST, &handle, size);

    last_bytes[id] = cinfo->nbytes;
    last_dropbytes[id] = cinfo->drop_bytes;

    rval = sizeof(struct class_info);
    return rval;
}

double get_timeindouble(void)
{
    double sec;
    static struct timeval start;

    if (packets_received == 0)
	start = remote_time;
    
    sec = (double)(remote_time.tv_sec - start.tv_sec)
	+ (double)(remote_time.tv_usec - start.tv_usec) / 1000000.0;
    return sec;
}

/* do endian conversion.  we assume all fields are 32 bits. */
static void endian_swap(char *buf, int size)
{
    long tmp, *p = (long *)buf;
    char *cp, *tp = (char *)&tmp;
    int i;
    
    cp = buf;
    for (i=0; i<size; i+=4) {
	tmp = *p++;
	cp[0] = tp[3]; cp[1] = tp[2]; cp[2] = tp[1]; cp[3] = tp[0];
	cp += 4;
    }
}

static void
sighup_handler(int sig)
{
	got_sighup = 1;
}


char *class_getname(long type, long *id)
{
    char *name = NULL;
    long handle = *id;

    if (handle < 0) {
	switch (handle) {
	case 0xffffffff:
	    name = "NULL";
	    break;
	case 0xfffffffe:
	    name = "root";
	    break;
	case 0xfffffffd:
	    name = "default";
	    break;
	case 0xfffffffc:
	    name = "control";
	    break;
	}
    }
    else if (handle < CBQM_MAXCLASS)
	name = class_name_tab[handle];

    if (name == NULL)
	name = "bogus";
    return (strdup(name));
}

/*
 * read config file and extract class names
 * currently, only works with the first interface that appears in the
 * config file.
 */
static int read_config(char *configfile)
{
    FILE		*fp;
    char	cmd_line[256], cmd_op[80];
    struct cmds *cmdp;
    char	*cp, *class_name;
    int		class_no;
	
    fp = fopen(configfile, "r");
    if (fp == NULL) 
	perror("can't open config file!");

    for (class_no = 0; class_no < CBQM_MAXCLASS; class_no++)
	if (class_name_tab[class_no] != NULL) {
	    free(class_name_tab[class_no]);
	    class_name_tab[class_no] = NULL;
	}
    
    class_no = 0;
    if (fp != NULL) while (1) {
	/* read a line from the config file */
	if (fgets(cmd_line, sizeof(cmd_line), fp) == NULL)
	    /* EOF */
	    break;
	
	/* remove trailing NL */
	cp = cmd_line + strlen(cmd_line) - 1;
	if (*cp == '\n')
	    *cp = '\0';
	else {
	    printf("LINE > 255 CHARS: %s.\n", cmd_line);
	    break;
	}

	if (cmd_line[0] == '#') 	/* Comment, skip this line */
	    continue;
	
	cp = cmd_line;
	if (!next_word(&cp, cmd_op))
		continue;

	if (strcmp(cmd_op, "class") == 0)
	    if ((class_name = cbq_class_cmd(cp)) != NULL)
		class_name_tab[class_no++] = class_name;
    };

    if (fp != NULL)
        (void) fclose(fp);

    for (; class_no < CBQM_MAXCLASS; class_no++) {
	sprintf(cmd_line, "class-%d", class_no);
	class_name_tab[class_no] = strdup(cmd_line);
    }
    
    return (0);
}

#define EQUAL(s1, s2)	(strcmp((char *)(s1), (char *)(s2)) == 0)

static char *
cbq_class_cmd(char *cmdbuf) 
{
    char	w[128], class_name[64], parent_name[64], *if_name;
    char	*cp = cmdbuf;

    /* get scheduling class. */
    if (!next_word(&cp, w))
	return (NULL);
    if (!EQUAL(w, "cbq"))
	return (NULL);

    /* get interface name. */
    if (!get_ifname(&cp, &if_name))
	return (NULL);

    /* get class name */
    if (!next_word(&cp, class_name))
	return (NULL);
    if (EQUAL(class_name, "ctl_class"))
	return (NULL);

    /* get parent name */
    if (!next_word(&cp, parent_name))
	return (NULL);
    if (EQUAL(parent_name, "null") || EQUAL(parent_name, "NULL"))
	return (NULL);

    /* check if this is a default class */
    while (next_word(&cp, w)) {
	if (EQUAL(w, "default"))
	    return (NULL);

	/* check # */
    }

    /* ok class name found */
    return (strdup(class_name));
}

/*
 * Skip leading blanks, then copy next word (delimited by blank or zero, but
 * no longer than 63 bytes) into buffer b, set scan pointer to following 
 * non-blank (or end of string), and return 1.  If there is no non-blank text,
 * set scan ptr to point to 0 byte and return 0.
 */
#ifndef MIN
#define MIN(a, b)	((a) < (b) ? (a) : (b))
#endif

static int 
next_word(char **cpp, char *b)
{
	char           *tp;
	int		L;

	*cpp += strspn(*cpp, " \t");
	if (**cpp == '\0' || **cpp == '\n' || **cpp == '#')
		return(0);

	tp = strpbrk(*cpp, " \t\n#");
	L = MIN((tp)?(tp-*cpp):strlen(*cpp), 63);
	strncpy(b, *cpp, L);
	*(b + L) = '\0';
	*cpp += L;
	*cpp += strspn(*cpp, " \t");
	return (1);
}


static int
get_ifname(char **cpp, char **ifnamep)
{
    char w[128], *ocp;
    static char interface[64];

    ocp = *cpp;
    if (!next_word(&ocp, w) || !isalpha(w[0]) || !isdigit(w[strlen(w)-1])) {
	/* this is not interface name. use one in the context. */
	*ifnamep = interface;
	if (interface[0] == 0)	/* no if_name saved */
	    return (0);
	return (1);
    }

    /* save the interface name */
    if (interface[0] == 0)
	strcpy(interface, w);
    /* check if this is not the second interface */
    if (strcmp(w, interface))
	return (0);

    *cpp = ocp;  /* advance the word pointer. */
    return (1);
}

