/* $Id: redstat.c,v 0.3 1998/04/03 07:35:46 kjc Exp kjc $ */
/*
 * Copyright (C) 1997
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <limits.h>
#include <errno.h>
#include <err.h>

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <net/if.h>

#include <netinet/altq_red.h>

#define	RED_DEVICE	"/dev/red"

int fd = -1;
char *if_name = NULL;
int interval = 5;
int avg_scale = 4096;	/* default fixed-point scale */

static void usage(void)
{
    fprintf(stderr, "usage: redstat [-i interval] [-s avg_scale] interface\n");
    exit(1);
}

static void
sigint_handler(int sig)
{
    fprintf(stderr, "Exiting on signal %d\n", sig);

    close(fd);

    exit(0);
}

int main(int argc, char **argv)
{
    struct red_stats red_stats;
    struct timeval cur_time, last_time;
    quad_t xmit_bytes, last_bytes;
    int msec;
    double kbps;
    char *if_name;
    int ch;

    while ((ch = getopt(argc, argv, "i:s:")) != EOF) {
	switch (ch) {
	case 'i':
	    interval = atoi(optarg);
	    break;
	case 's':
	    avg_scale = atoi(optarg);
	    break;
	}
    }
    argc -= optind;
    argv += optind;
    if (argc > 0)
	if_name = argv[0];
    else {
	if (argc == 0)
	    fprintf(stderr, "missing interface name\n");
	usage();
    }

    signal(SIGINT, sigint_handler);
    signal(SIGTERM, sigint_handler);
    
    if ((fd = open(RED_DEVICE, O_RDONLY)) < 0)
	err(1, "red open");

    if (strlen(if_name) > IFNAMSIZ)
	err(1, "interface name %s too long!\n", if_name);
    red_stats.iface.red_ifname[IFNAMSIZ-1] = '\0';
    strncpy(red_stats.iface.red_ifname, if_name, IFNAMSIZ-1);

    gettimeofday(&last_time, NULL);
    last_time.tv_sec -= interval;
    last_bytes = 0;

    while (1) {
	
	if (ioctl(fd, RED_GETSTATS, &red_stats) < 0)
	    err(1, "ioctl RED_GETSTATS");

	gettimeofday(&cur_time, NULL);
	msec = (cur_time.tv_sec - last_time.tv_sec)*1000 +
	    (cur_time.tv_usec - last_time.tv_usec)/1000;

	/*
	 * measure the throughput of this class
	 */
	xmit_bytes = red_stats.xmit_bytes - last_bytes;
	if (xmit_bytes < 0)
	    xmit_bytes += QUAD_MAX;
	kbps = (double)xmit_bytes * 8.0 / (double)msec * 1000.0 / 1000.0;
	last_bytes = red_stats.xmit_bytes;
	last_time = cur_time;
	
	printf(" q_len:%d (avg: %.2f), q_limit:%d\n",
	       red_stats.q_len,  ((double)red_stats.q_avg)/(double)avg_scale,
	       red_stats.q_limit);
	printf(" xmit: %qd pkts, drop: %qd pkts (forced: %qd, early: %qd)\n",
	       red_stats.xmit_packets, red_stats.drop_packets,
	       red_stats.drop_forced, red_stats.drop_unforced);
	if (red_stats.marked_packets != 0)
		printf(" marked: %qd\n", red_stats.marked_packets);
	if (kbps > 1000.0)
	    printf(" throughput: %.2f Mbps\n", kbps/1000.0);
	else
	    printf(" throughput: %.2f Kbps\n", kbps);

	sleep(interval);
    }

    return (0);
}
