/*
 * Copyright (C) 1997
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: wfqd.c,v 0.2 1998/04/03 07:36:28 kjc Exp kjc $
 */
/*
 *  March 27, 1997.  Written by Hiroshi Kyusojin of Keio University
 *  (kyu@mt.cs.keio.ac.jp). 
 */

#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<string.h>
#include<errno.h>
#include<ctype.h>
#include<sys/socket.h>
#include<sys/ioctl.h>
#include<sys/fcntl.h>
#include<netinet/in.h>
#include<arpa/inet.h>
#include<netinet/altq_wfq.h>

#define DEBUG
#define HASH_BY_DST_ADDR

#define WFQ_DEV		"/dev/wfq"
#ifndef IFNAMSIZ
#define IFNAMSIZ	16
#endif
#define PROMPT		"> "
#define QUIT		"quit"
#define ENABLE		"enable"
#define DISABLE		"disable"
#define GETSTATS	"getstats"
#define SETWEIGHT	"setweight"
#ifdef DEBUG
#define GETQID		"getqid"
#endif

#define ERROR(format, args...)				\
	  do{						\
		fprintf(stderr, format, ##args);	\
		exit(1);				\
	  }while(0)
#define PERROR(msg)					\
	  do{						\
		perror(msg);				\
		exit(1);				\
	  }while(0)

/*
 * prototype
 */
int get_ifname(char *, char *);
int do_command(int, char *);

void main(int argc, char *argv[])
{
  int fd;
  char buf[256];

  /*
   * open file and get a file descriptor
   */
  if((fd = open(WFQ_DEV, O_RDONLY)) == NULL)
    PERROR("Can't open wfq device file");

  while (--argc > 0) {
      ++argv;
      if (argv[0][0] == '-') {
	  /* add options */
	  printf("usage: wfqd [if_name]\n");
	  exit(1);
      }
      else {
	/* must be interface name. enable interface. */

	if (argc > 1 && isdigit(argv[1][0])) {
	  sprintf(buf, "%s %s %s", argv[0], ENABLE, argv[1]);
	  --argc; ++argv;
	}
	else
	  sprintf(buf, "%s %s", argv[0], ENABLE);

	do_command(fd, buf);
      }
  }

  while(1){
    printf("%s", PROMPT);
    fgets(buf, sizeof(buf), stdin);
    
    /*
     * execute command
     */
    if (do_command(fd, buf) < 0)
      break;
  }
  

  /*
   * close a file descriptor
   */
  close(fd);
}


int get_ifname(char *ifname, char *p)
{
  int i;

  for(i = 0; i < IFNAMSIZ - 1 && *p != ' ' && *p != '\t' && *p != '\n' ; i++)
    ifname[i] = *p++;
  ifname[i] = '\0';

  return i;
}


int do_command(int fd, char *buf)
{
  struct wfq_interface iface;
  int len;
  char *p, ifname[IFNAMSIZ];

  for(p = buf; *p == ' ' || *p == '\t'; p++)
    ;
  if(*p == '\n')
    return 0;

  if(! strncmp(QUIT, p, strlen(QUIT)))
    return -1;

  /*
   * get interface name
   */
  if((len = get_ifname(ifname, p)) == IFNAMSIZ - 1){
    fprintf(stderr, "too long interface name : %s\n", ifname);
    return 0;
  }
  strcpy(iface.wfq_ifacename, ifname);
  iface.wfq_ifacelen = len;

  p += len;
  while(*p == ' ' || *p == '\t')
    p++;

  if(! strncmp(ENABLE, p, strlen(ENABLE))){
    p += strlen(ENABLE);
    while(*p == ' ' || *p == '\t')
      p++;
    for(iface.nqueues = 0; '0' <= *p && *p <= '9'; p++)
      iface.nqueues = iface.nqueues * 10 + *p - '0';
    if(ioctl(fd, WFQ_IF_ATTACH, &iface) < 0)
      PERROR("ioctl WFQ_IF_ATTACH");
    if(ioctl(fd, WFQ_ENABLE, &iface) < 0)
      PERROR("ioctl WFQ_ENABLE");
  }
  
  else if(! strncmp(DISABLE, p, strlen(DISABLE))){
    if(ioctl(fd, WFQ_DISABLE, &iface) < 0)
      PERROR("ioctl WFQ_DISABLE");
    if(ioctl(fd, WFQ_IF_DETACH, &iface) < 0)
      PERROR("ioctl WFQ_IF_DETACH");
  }

#ifdef HASH_BY_DST_ADDR
  else if(! strncmp(GETQID, p, strlen(GETQID))){
    struct wfq_getqid gqid;
    struct flowinfo_in *fi;
    u_long tmp, val = 0;

    bzero(&gqid, sizeof(gqid));
    gqid.iface = iface;

    p += strlen(GETQID);
    while(*p == ' ' || *p == '\t')
      p++;

    do{
      for(tmp = 0;  '0' <= *p && *p <= '9'; p++)
	tmp = tmp * 10 + *p - '0';
      val = (val << 8) | tmp;
    }while(*p++ == '.');
    fi = (struct flowinfo_in *)&gqid.flow;
    fi->fi_family = AF_INET;
    fi->fi_dst.s_addr = htonl(val);
    if(ioctl(fd, WFQ_GET_QID, &gqid) < 0)
      PERROR("ioctl WFQ_GET_QID");

    printf("queue ID = %ld for dest (%s)\n",
	   gqid.qid, inet_ntoa(fi->fi_dst));
  }
#endif

  else if(! strncmp(SETWEIGHT, p, strlen(SETWEIGHT))){
    struct wfq_setweight sw;
    
    sw.iface = iface;
    
    p += strlen(SETWEIGHT);
    while(*p == ' ' || *p == '\t')
      p++;
    
    for(sw.qid = 0; '0' <= *p && *p <= '9'; p++)
      sw.qid = sw.qid * 10 + *p - '0';
    
    while(*p == ' ' || *p == '\t')
      p++;
    
    for(sw.weight = 0; '0' <= *p && *p <= '9'; p++)
      sw.weight = sw.weight * 10 + *p - '0';
    if(ioctl(fd, WFQ_SET_WEIGHT, &sw) < 0)
      PERROR("ioctl WFQ_SET_WEIGHT");
  }

  else if(! strncmp(GETSTATS, p, strlen(GETSTATS))){
    struct wfq_getstats gs;
    
    gs.iface = iface;
    
    p += strlen(GETSTATS);
    while(*p == ' ' || *p == '\t')
      p++;
    
    for(gs.qid = 0; '0' <= *p && *p <= '9'; p++)
      gs.qid = gs.qid * 10 + *p - '0';

    if(ioctl(fd, WFQ_GET_STATS, &gs) < 0)
      PERROR("ioctl WFQ_GET_STATS");
    
    printf("Bytes  = %d\n", gs.stats.bytes);
    printf("Weight = %d\n", gs.stats.weight);
    printf("Sent : packets = %d, bytes = %d\n",
	   gs.stats.sent_packets, gs.stats.sent_bytes);
    printf("Drop : packets = %d, bytes = %d\n",
	   gs.stats.drop_packets, gs.stats.drop_bytes);
  }

  else{
    fprintf(stderr, "Usage : interface %s [nqueues]\n", ENABLE);
    fprintf(stderr, "        interface %s\n", DISABLE);
#ifdef HASH_BY_DST_ADDR
    fprintf(stderr, "        interface %s ipaddr(XX.XX.XX.XX)\n", GETQID);
#endif
    fprintf(stderr, "        interface %s qid\n", GETSTATS);
    fprintf(stderr, "        interface %s qid weight\n", SETWEIGHT);
  }

  return 0;
}
  
